use ir::{CrudField, CrudFieldAttributes, CrudStructAttributes};
use quote::quote;
use syn::DeriveInput;

mod ir;

fn extract_crud_field_attrs(ast: &mut DeriveInput) -> deluxe::Result<(Vec<CrudField>, CrudField)> {
    let mut field_attrs: Vec<CrudField> = Vec::new();
    // let mut identifier: Option<CrudIdentifier> = None;
    let mut identifier: Option<CrudField> = None;
    let mut identifier_counter = 0;
    if let syn::Data::Struct(s) = &mut ast.data {
        for field in &mut s.fields {
            let ident = field.clone().ident.unwrap();
            let ty = field.clone().ty;
            let attrs: CrudFieldAttributes =
                deluxe::extract_attributes(field).expect("Could not extract attributes from field");
            let field = CrudField {
                ident: ident.clone(),
                field: field.to_owned(),
                column: attrs.column.unwrap_or_else(|| ident.to_string()),
                id: attrs.id,
                ty,
            };
            if attrs.id {
                identifier_counter += 1;
                identifier = Some(field.clone());
            }
            if identifier_counter > 1 {
                return Err(syn::Error::new_spanned(
                    field.field,
                    "Struct {name} can only have one identifier",
                ));
            }
            field_attrs.push(field);
        }
    }
    if identifier_counter < 1 {
        Err(syn::Error::new_spanned(
            ast,
            "Struct {name} must have one identifier",
        ))
    } else {
        Ok((field_attrs, identifier.unwrap()))
    }
}

fn generate_find_query(table: &str, id: &CrudField) -> proc_macro2::TokenStream {
    let find_string = format!("SELECT * FROM {} WHERE {} = $1", table, id.column);
    let ty = &id.ty;
    quote! {
        async fn find(pool: &::sqlx::PgPool, id: &#ty) -> ::sqlx::Result<Option<Self>> {
            ::sqlx::query_as!(Self, #find_string, id)
                .fetch_optional(pool)
                .await
        }
    }
}

fn generate_create_query(table: &str, fields: &[CrudField]) -> proc_macro2::TokenStream {
    let inputs: Vec<String> = (1..=fields.len()).map(|num| format!("${num}")).collect();
    let create_string = format!(
        "INSERT INTO {} ({}) VALUES ({}) RETURNING *",
        table,
        fields
            .iter()
            .map(|v| v.column.clone())
            .collect::<Vec<String>>()
            .join(", "),
        inputs.join(", ")
    );
    let field_idents: Vec<syn::Ident> = fields.iter().map(|f| f.ident.clone()).collect();
    quote! {
        async fn create(&self, pool: &::sqlx::PgPool) -> ::sqlx::Result<Self> {
            ::sqlx::query_as!(
                Self,
                #create_string,
                #(self.#field_idents),*
            )
            .fetch_one(pool)
            .await
        }
    }
}

fn generate_update_query(
    table: &str,
    fields: &[CrudField],
    id: &CrudField,
) -> proc_macro2::TokenStream {
    let mut fields: Vec<&CrudField> = fields.iter().filter(|f| !f.id).collect();
    let update_columns = fields
        .iter()
        .enumerate()
        .map(|(i, &field)| format!("{} = ${}", field.column, i + 1))
        .collect::<Vec<String>>()
        .join(", ");
    let update_string = format!(
        "UPDATE {} SET {} WHERE {} = ${} RETURNING *",
        table,
        update_columns,
        id.column,
        fields.len() + 1
    );
    fields.push(id);
    let field_idents: Vec<_> = fields.iter().map(|f| f.ident.clone()).collect();
    quote! {
        async fn update(&self, pool: &::sqlx::PgPool) -> ::sqlx::Result<Self> {
            ::sqlx::query_as!(
                Self,
                #update_string,
                #(self.#field_idents),*
            )
            .fetch_one(pool)
            .await
        }
    }
}

fn generate_delete_query(table: &str, id: &CrudField) -> proc_macro2::TokenStream {
    let delete_string = format!("DELETE FROM {} WHERE {} = $1", table, id.column);
    let ty = &id.ty;
    let ident = &id.ident;

    quote! {
        async fn delete_by_id(pool: &::sqlx::PgPool, id: &#ty) -> ::sqlx::Result<u64> {
            let rows_affected = ::sqlx::query!(#delete_string, id)
                .execute(pool)
                .await?
                .rows_affected();
            Ok(rows_affected)
        }

        async fn delete(&self, pool: &::sqlx::PgPool) -> ::sqlx::Result<u64> {
            let rows_affected = ::sqlx::query!(#delete_string, self.#ident)
                .execute(pool)
                .await?
                .rows_affected();
            Ok(rows_affected)
        }
    }
}

pub fn crud_derive_macro2(
    item: proc_macro2::TokenStream,
) -> deluxe::Result<proc_macro2::TokenStream> {
    // parse
    let mut ast: DeriveInput = syn::parse2(item).expect("Failed to parse input");

    // extract struct attributes
    let CrudStructAttributes { table } =
        deluxe::extract_attributes(&mut ast).expect("Could not extract attributes from struct");

    // extract field attributes
    let (fields, id) = extract_crud_field_attrs(&mut ast)?;
    let ty = &id.ty;
    let id_ident = &id.ident;

    // define impl variables
    let ident = &ast.ident;
    let (impl_generics, type_generics, where_clause) = ast.generics.split_for_impl();

    // generate
    let find_query = generate_find_query(&table, &id);
    let create_query = generate_create_query(&table, &fields);
    let update_query = generate_update_query(&table, &fields, &id);
    let delete_query = generate_delete_query(&table, &id);
    let code = quote! {
        impl #impl_generics Crud<#ty> for #ident #type_generics #where_clause {
            #find_query

            #create_query

            #update_query

            async fn create_or_update(&self, pool: &::sqlx::PgPool) -> ::sqlx::Result<Self> {
                if Self::find(pool, &self.#id_ident).await?.is_some() {
                    self.update(pool).await
                } else {
                    self.create(pool).await
                }
            }

            #delete_query
        }
    };
    Ok(code)
}
